/*******************************************************
*
*	SM3D Camera definition
*
********************************************************
*	SM3D
*
*	File type	:	Source CPP
*	Author		:	Dario Deledda
*	Date		:	25/11/2000
*	Last update	:	25/11/2000
*	Version		:	0.9a
*
********************************************************/
#include "../header/SM3DCamera.h"

// Variabili statiche delle luci
smPointC		*lighta=0;		// puntatore al buffer delle luci
UInt32	nlight=0;				// numero luci attive
#define	max_nlight	10		// massimo 10 luci per scena
//	t valore
//	s distanza massima

void set_Light(UInt32 n,smPointC &in){
	lighta[n]=in;
}

void set_NumLight(UInt32 n){
	nlight=n;
}

void init_Light(void){
	if(!lighta){
		lighta=new  smPointC[max_nlight];	// alloca la memoria le luci
	}
}

void del_Light(void){
	if(lighta){
		delete[] lighta;
	}
}

SMErr calc_Light(const smPointC &pin,const smPointC &n,smRGB &ret){
	register Int32 t=nlight;		// t fa da counter per le luci presenti
	register Fl32 r,g,b,vl,m;
	register smPointC l;
	register smPointC p1,p2,p3;
	register Fl32 maxv=max_light;
	
	r=0; g=0; b=0;
	while (t>0){
		l=*(lighta+(t-1));
		p1=pin;
		p2=n;
		
		p1=l-pin;
		if(abs(p1.x)<l.s && abs(p1.y)<l.s && abs(p1.z)<l.s){
			m=p1.mod();
			if (m<=l.s) {	// se la luce e' nel raggio di azione esegui
				p1=p1/m;
				
				vl=(p1*p2);
				//if (vl <0) vl=-vl;	// per luci false 
				if(vl<0) vl=0;
				
				vl=vl*l.t;
				vl=vl*(1-(m/l.s));
				
				//if(!Cnegatf){
				//	if(vl<0) vl=0;
				//}
				if (vl<-maxv) vl=-maxv;
				else
				if(vl>maxv) vl=maxv;

				m=vl/maxv;
				
				r=r+l.cl.r*m;
				g=g+l.cl.g*m;
				b=b+l.cl.b*m;
			}
		}
		--t;
	}
	
/*	maxv=2*maxv;
	if (r>maxv) r=maxv;
	if (r<maxv) r=-maxv;
	if (g>maxv) g=maxv;
	if (g<maxv) g=-maxv;
	if (b>maxv) b=maxv;
	if (b<maxv) b=-maxv;
*/
	ret.r=r-127; ret.g=g-127; ret.b=b-127;
	ret.r<<=3;
	ret.g<<=3;
	ret.b<<=3;
	//ret.r+=ar; ret.g+=ag; ret.b+=ab;
	return SM3DErr_NoErr;
}

SMErr	smCamera::CalcLightPoly(smPointC *inp, UInt16 num){
	UInt16 cnt;
	smPointC norm;
	smRGB outc;
	norm=((inp[num-1]-inp[0])/(inp[1]-inp[0]));
	for(cnt=0; cnt<num; cnt++){
		 calc_Light(inp[cnt],norm,outc);
		 inp[cnt].cl.r=outc.r;
		 inp[cnt].cl.g=outc.g;
		 inp[cnt].cl.b=outc.b;	 
	}
	return SM3DErr_NoErr;
}

smViewPort::smViewPort(){
	xclip=smHWVwidth/2,yclip=smHWVheight/2;		// Valori di clip orizontali e verticali

	xc_v=smHWVwidth/2,yc_v=smHWVheight/2;		// xc,yc meta' dimensione video visibile

	xclip_l=0,xclip_r=smHWVwidth;		// rettangolo visualizzato a video
	yclip_u=0,yclip_b=smHWVheight;		// xclip_l,yclip_u sono le coordinate spigolo alto sinistra
										// xclip_r la larghezza,yclip_b l' altezza

	xclip_cl=-xc_v,xclip_cr=xc_v;		//  coordinate di cliping per la x
	yclip_cu=-yc_v,yclip_cb=yc_v;		//  coordinate di cliping per la y
}

SMErr smViewPort::SetViewCenter(Int16 xc,Int16 yc){
	xclip=xc;
	yclip=yc;
	return SM3DErr_NoErr;
}

SMErr smViewPort::SetViewRect(Int16 xl,Int16 yu,Int16 xr,Int16 yb){
	if (xl >= xr || yu >= yb) 		return SM3DErr_No_Clip_Rect;	// non e' un rettangolo di cliping
	if (xl > smHWVwidth || xr <  0) return SM3DErr_Out_Of_Video; 	// rettangolo fuori video
	if (yu > smHWVheight || yb < 0) return SM3DErr_Out_Of_Video;	// rettangolo fuori video
	
	if (xl < 0) xl=0;					
	if (xr >smHWVwidth) xl=smHWVwidth;
	if (yu < 0) yu=0;
	if (yb > smHWVheight) yb=smHWVheight;
	
	xclip_l=xl; xclip_r=xr;
	yclip_u=yu; yclip_b=yb;

	xclip_cl=(xl-xc_v); xclip_cr=(xr-xc_v);	
	yclip_cu=(yu-yc_v);yclip_cb=(yb-yc_v);
	
	return SM3DErr_NoErr;
}

void smCamera::ReinitDistSc(){
	register Int32 t,t1;
	for (t=-xclip; t<=xclip; t++)		// calcolo preliminare della m per la x
		{t1=t; mcx[xc_v+t]=(t1<<Ks_b)/dist_sc;}
		
	for (t=-yclip; t<=yclip; t++)		// calcolo preliminare per la n per la y
		{t1=t; mcy[yc_v-t]=(t1<<Ks_b)/dist_sc;}
}

smPoint32 smCamera::TreToDued(register const smPoint32& punto_in,register SMErr& er){	
	register Int32	cs,sn;
	register smPoint32 pt,punto_out;
	register Fl32 trash;
	er=0;
	
	pt.x=punto_in.x-x;
	pt.y=punto_in.y-y;
	pt.z=punto_in.z-z;

	if (ang.x !=0) {
		cs=fcos[(Int16)ang.x];
		sn=fsin[(Int16)ang.x];
		punto_out=pt;
		pt.y=(punto_out.y*cs+punto_out.z*sn)>>Ks_b;
		pt.z=(-punto_out.y*sn+punto_out.z*cs)>>Ks_b;
	}
	if (ang.y !=0) {
		cs=fcos[(Int16)ang.y];
		sn=fsin[(Int16)ang.y];
		punto_out=pt;
		pt.z=(punto_out.z*cs+punto_out.x*sn)>>Ks_b;
		pt.x=(-punto_out.z*sn+punto_out.x*cs)>>Ks_b;
	}
	if (ang.z !=0) {
		cs=fcos[(Int16)ang.z];
		sn=fsin[(Int16)ang.z];
		punto_out=pt;
		pt.y=(punto_out.y*cs+punto_out.x*sn)>>Ks_b;
		pt.x=(-punto_out.y*sn+punto_out.x*cs)>>Ks_b;
	}
//*/
	if (pt.z <= 0) er=SM3DErr_Out_Of_Video;
	else {
		trash=(Fl32)dist_sc/(Fl32)pt.z;
		punto_out.y=(Fl32)pt.y * trash;
		punto_out.x=(Fl32)pt.x * trash;
		punto_out.z=pt.z;
		er=SM3DErr_NoErr;	
	}
	return punto_out;
}

smPointC smCamera::Duef(register smPointC &punto_in,SMErr& er)
{
	register smPointC punto_out=punto_in;
	register Fl32 trash;

	if (punto_in.z >0){
		trash=(Fl32)dist_sc/punto_in.z;
		punto_out.y=(Fl32)punto_in.y * trash;
		punto_out.x=(Fl32)punto_in.x * trash;
		punto_out.z=0;
		er=SM3DErr_NoErr;	
	}
	else
		er=SM3DErr_Out_Of_Video;

	return punto_out;
}
//	Routine di clipping della classe _camerav
//inline
 bool smCamera::Clip(register smPointC& p1,register smPointC& p2){
 	Fl32 dy,dx,ds,dt,bf,bz;
 	Fl32 dr,dg,db,da;
// 	Fl32 x,y,z,s,t;
	const Fl32 zpc=35.0f;

 	if (p1.z <zpc && p2.z <zpc) return false;

	if (p1.z != p2.z) {
		if(p1.z <zpc){
			bf=1.f/(p2.z-p1.z);
			dx=(Fl32)(p2.x-p1.x)*bf;
			dy=(Fl32)(p2.y-p1.y)*bf;
			ds=(Fl32)(p2.s-p1.s)*bf;
			dt=(Fl32)(p2.t-p1.t)*bf;

			dr=(Fl32)(p2.cl.r-p1.cl.r)*bf;
			dg=(Fl32)(p2.cl.g-p1.cl.g)*bf;
			db=(Fl32)(p2.cl.b-p1.cl.b)*bf;
			da=(Fl32)(p2.cl.a-p1.cl.a)*bf;
			bz=(zpc-p1.z);
			
			p1.x=dx*bz+(Fl32)p1.x;
			p1.y=dy*bz+(Fl32)p1.y;
			p1.z=zpc;
			p1.t=dt*bz+(Fl32)p1.t;
			p1.s=ds*bz+(Fl32)p1.s;
			
			p1.cl.r=dr*bz+(Fl32)p1.cl.r;
			p1.cl.g=dg*bz+(Fl32)p1.cl.g;
			p1.cl.b=db*bz+(Fl32)p1.cl.b;
			p1.cl.a=da*bz+(Fl32)p1.cl.a;
			
			return true;
		} else
		if(p2.z<zpc){
			bf=1.f/(p1.z-p2.z);
			dx=(Fl32)(p1.x-p2.x)*bf;
			dy=(Fl32)(p1.y-p2.y)*bf;
			ds=(Fl32)(p1.s-p2.s)*bf;
			dt=(Fl32)(p1.t-p2.t)*bf;

			dr=(Fl32)(p1.cl.r-p2.cl.r)*bf;
			dg=(Fl32)(p1.cl.g-p2.cl.g)*bf;
			db=(Fl32)(p1.cl.b-p2.cl.b)*bf;
			da=(Fl32)(p1.cl.a-p2.cl.a)*bf;
			bz=(zpc-p2.z);
			
			p2.x=dx*bz+(Fl32)p2.x;
			p2.y=dy*bz+(Fl32)p2.y;
			p2.z=zpc;
			p2.t=dt*bz+(Fl32)p2.t;
			p2.s=ds*bz+(Fl32)p2.s;
			
			p2.cl.r=dr*bz+(Fl32)p2.cl.r;
			p2.cl.g=dg*bz+(Fl32)p2.cl.g;
			p2.cl.b=db*bz+(Fl32)p2.cl.b;
			p2.cl.a=da*bz+(Fl32)p2.cl.a;
			
			return true;
		}
	}	
	return true;
}

void smCamera::insert2(register smRGB &x,register Int32 y){
	x.v+=xc_v;	// coordinata x del punto inserito
	x.n+=xc_v;	// verra usato per il confronto se lo spigolo sinistro
	if ( x.v<xclip_l) {	// e' fuori schermo
		x.v=xclip_l; 		
	} else if (x.v>=xclip_r) {
		x.v=xclip_r-1;
	}
	if (y>yclip_u && y<yclip_b){
		if (y<ymin) ymin=y; else if (y>ymax) ymax=y;
		
		if (pixp[y] != pixpc) {
			px[y][0]=x;
			pix[y]=0;
			pixp[y]=pixpc; 
			pix[y]=1; return;
		}else if (pix[y]==1) {
			if (x.v<px[y][0].v) {	
				px[y][1]=px[y][0];
				 px[y][0]=x; ++pix[y];
			}else {
				px[y][1]=x; ++pix[y];
			}
		}
		else {
			if (x.v < px[y][0].v) {px[y][0]=x;} 
			else if (x.v>px[y][1].v) {px[y][1]=x;}
		}	

	}
	return;
}

SMErr smCamera::smTriangle(smPointC &p0i,smPointC &p1i,smPointC &p2i,smTexture *texture,bool twoFace){
	smPoint pnt[3];
	pnt[0]=p0i;
	pnt[1]=p1i;
	pnt[2]=p2i;

	return smQuad(pnt,3,texture,true);	
}

SMErr smCamera::smQuad(smPointC *pnt,UInt16 num,smTexture *texture,bool twoFace){

	register smPointC	pnti[12],pnto[12],pnto2[24];
	register smPointC	p1,p2,tmp;	
	register smPointC	pt1,pt2;
	smPointC			P,N,M,norm;
	
	register Int32		t;
	register Int32		cs,sn;
	register Int32		p_count=0,pv_c=0;
	Int32 				clip_b;	
	
	register Fl32		cd;
	bool				v=false;
	SMErr				er1=0,er2=0;
	smRGB				c1,c2;

	Int32			np=num+1;
	
	for(t=0; t<num; t++)
		pnti[t]=pnt[t];
	
	pnti[t]=pnt[0];			// close the last vertex
		
	++pixpc;				// incremento per l'array pix[]
	
	ymin=smHWVheight;
	ymax=0;

	clip_b=0;
	p2.x=this->x;
	p2.y=this->y;
	p2.z=this->z;
		
	for(p_count=0; p_count <np;p_count++){

		p1=pnti[p_count];
		p1=p1-p2;
		
		// rotation	
		if (ang.y !=0) {
			cs=cosv[Ncmp-ang.y];
			sn=sinv[Ncmp-ang.y];
			tmp=p1;
			p1.z=(tmp.z*cs+tmp.x*sn)>>Ks_b;
			p1.x=(-tmp.z*sn+tmp.x*cs)>>Ks_b;
		}
		if (ang.x !=0) {
			cs=cosv[Ncmp-ang.x];
			sn=sinv[Ncmp-ang.x];
			tmp=p1;
			p1.y=(tmp.y*cs+tmp.z*sn)>>Ks_b;
			p1.z=(-tmp.y*sn+tmp.z*cs)>>Ks_b;
		}
		if (ang.z !=0) {
			cs=cosv[Ncmp-ang.z];
			sn=sinv[Ncmp-ang.z];
			tmp=p1;
			p1.y=(tmp.y*cs+tmp.x*sn)>>Ks_b;
			p1.x=(-tmp.y*sn+tmp.x*cs)>>Ks_b;
		}
//*/		
		pnto[p_count]=p1;
		if (p1.z <= 0) ++clip_b;	// if the point have z<0 inc clip counter		
	}	
	
	// if the polygon is out of video return an error
	if(clip_b == np+1) return SM3DErr_Out_Of_Video;

	// calculate the normal to the triangle
	P=pnto[0];
	M=pnto[1]; M=M-P; 
	N=pnto[num-1]; N=N-P;
	
	norm=N/M;

	//cd=-(norm*P);
	cd=-(norm.x*P.x+norm.y*P.y+norm.z*P.z);
	if (!twoFace && (cd > 0)) return SM3DErr_No_Visible;
	
	for(p_count=0,t=0 ; p_count<np ; ){
		p1=pnto[p_count];	
		p2=pnto[++p_count];	

		v=false;
		v=Clip(p1,p2);		// clipping con solo piano z
		//v=Clip1(p1,p2);	// clipping con tronco di piramide
		if (v){
			pnti[t++]=p1;
			pnti[t++]=p2;
		} 
	}

	np=t-1;
	
	// if no visible point exit
	if (np<3) return SM3DErr_No_Visible;

		{
		register Int32	xl=0,xr=0,yh=0,yl=0;	// contatori per punti fuori video
		register Int32 	npt=np+1,nm=0;	// numero punti effettivi;
		p2=(long)0;	
		for(p_count=0; p_count<=np;){
			p1=pnti[p_count];			// p1 prende il punto da trasformare in 2d 
			if ( p1 != p2) {			// se p1 != p2 esegui il calcolo
				pt1=this->Duef(p1,er1);	// trasforma il punto p1 in 2d	
				pt1.z=er1;				// prendi anche il codice di errore
			}
			pnto2[p_count++]=pt1;
			p2=p1;
			
			if(er1) ++t; 
			else	
			if (pt1.x < xclip_cl)	++xl;		// fuori da video a sinistra
			else
			if (pt1.x > xclip_cr)	++xr;		// fuori dal video a destra
			else
			if (pt1.y < -yclip_cb) ++yl;		// fuori dal video sotto
			else
			if (pt1.y > -yclip_cu)	++yh;	// fuori dal video sopra
			else
			++nm;
		}
		if(nm==0){
			if (xl == npt) return SM3DErr_Out_Of_Video;	// poligono tutto a sinistra
			if (xr == npt) return SM3DErr_Out_Of_Video;	// poligono tutto a destra
			if (yl == npt) return SM3DErr_Out_Of_Video;	// poligono tutto sotto
			if (yh == npt) return SM3DErr_Out_Of_Video;	// poligono tutto alto		
			if (t == npt)  return SM3DErr_Out_Of_Video;	// poligono tutto dietro all' osservatore
		}
	}

	np--;
	for(p_count=0; p_count<np;){
		p1=pnti[p_count];
		pt1=pnto2[p_count];
		pt2=pnto2[++p_count];
		p2=pnti[p_count];
		er1=pt1.z;
		er2=pt2.z;

//		buf->smLine(pt1.x+80,64-pt1.y,pt2.x+80,64-pt2.y,smRGB16(0x7f0f));

		if(pt1!=pt2){
			c1=p1.cl;	// prende i colori dei 2 punti 
			c2=p2.cl;	//
			register Fl32 dif,mx,n,mxs;
			register Fl32 mr,mg,mb,mrs,mgs,mbs;
			register Fl32 ma,mas;
			register Int32 yp;
			register Int32 z1,z2;
			register smRGB a; 
			
			c1.CutCol();
			c2.CutCol();
			z1=p1.z;
			z2=p2.z;

			if (pt1.y>pt2.y) {smPointC tmpv=pt1; pt1=pt2; pt2=tmpv;
						 dif=z1;z1=z2;z2=dif;
						 a=c1; c1=c2; c2=a; }

		
			dif=(pt2.y-pt1.y);
			if (dif !=0 ){		
				mx=(pt2.x-pt1.x)/dif;	// delta per la x
				mr=(c2.r-c1.r)/dif;	// delta componente r colore
				mg=(c2.g-c1.g)/dif;	// delta componente g colore
				mb=(c2.b-c1.b)/dif;	// delta componente b colore
				ma=(c2.a-c1.a)/dif;	// delta componente b colore

				mxs=pt1.x-mx;
				mrs=c1.r-mr;
				mgs=c1.g-mg;
				mbs=c1.b-mb;
				mas=c1.a-ma;
				n=-1;
				yp=pt1.y;
				
				yp=yc_v-yp;
				
				if (pt1.y > yclip_b){		// punto sotto il video salta 
					register float t;	// i punti inutili
					t=(pt1.y-yclip_b);
					mxs+=mx*t;
					mrs+=mr*t;
					mgs+=mg*t;
					mbs+=mb*t;
					mas+=ma*t;
					n=t;--n;
					yp-=t;
				}
				while(++n <= dif && --yp>yclip_u){	//** modifica ++n < dif
					mxs+=mx;		// valore di x
					mrs+=mr;		//
					mgs+=mg;		// colore interpolato linearmente
					mbs+=mb;		//
					mas+=ma;		// canale alfa
					
					a.v=(Int32)mxs;
					
					a.n=a.v;
					
					a.r=mrs;
					a.g=mgs;
					a.b=mbs;
					a.a=mas;		// canale alfa
					insert2(a,yp);
					
				}
			} 
			else if (pt1.y>yclip_u && pt1.y<yclip_b)
				 { 	a=c1;a.v=pt1.x;a.n=pt1.x;
					insert2(a,yc_v-pt1.y);
					a=c2;a.v=pt2.x;a.n=pt2.x;
					insert2(a,yc_v-pt2.y);
				}
				
			
		}
	}
	{
		smRGB	p1,p2;
		register Fl32	d;
		register UInt32 *zpun=z_buf;		// da riattivare quando usati
		register UInt32 framet=frame1;
		register UInt16	*p8,*p9,*p10;
		register Int32 u1,v1;
		register Int32 dat,dat2;
		register Int32 dat1;
		
		register Int32 ri1,ri2,ri;	// rosso di partenza e di arrivo
		register Int32 gi1,gi2,gi;	// verde di partenza e di arrivo
		register Int32 bi1,bi2,bi;	// blu di partenza e di arrivo
		Int32 rs,bs,gs;
		Int32 y,x,u,t;
		Int32 mod1,mod2;
		register Int32 dxt,dyt,dzt;
		
		register Int32 xv,yv;

		register Fl32 a,b,c,h;		
		register Fl32 Ha1,Hb1,Va,Vb,Vc;
		Fl32 Oa,Ob,Oc,Ha,Hb,Hc;
		
		register UInt32 mask=0x1f;
		register UInt16 *colr=&col16r[col_tab];	//
		register UInt16 *colg=&col16g[col_tab];	// puntatori alle tabelle per il
		register UInt16 *colb=&col16b[col_tab];	// calcolo dell' illuminazione

		register Int32 dzb_,zb_;

		p10=(UInt16*)(texture->txt);		//p10 punta alla texture
		dxt=(texture->dx);				// dxt contiene la lunghezza della texture
		dyt=(texture->dy);				// dyt contiene l' altezza della texture
		dzt=(texture->sh);				// shift per la gestione offset texture
		--dxt;--dyt;					// diminuiti di 1 per l' AND logico es. 64->63
		
		
		mod1=(texture->dx * texture->lx);				// larghezza della texture
		mod2=(texture->dy * texture->ly)*texture->dy;	// altezza della texture

		dyt<<=dzt;		// nuova maschera per dyt con shift di dzt bit

		p8=(UInt16*)buf->GetBackBuffer();
		
		yv=ymin;
		d=(this->dist_sc);
						
        Oc = d*(M.y*N.x - M.x*N.y);
        Hc = M.z*N.y - M.y*N.z;
        Vc = M.x*N.z - M.z*N.x;

        Oa = d*(N.x*P.y - N.y*P.x);
        Ha = N.y*P.z - N.z*P.y;
        Va = N.z*P.x - N.x*P.z;

        Ob = d*(M.x*P.y - M.y*P.x);
        Hb = M.y*P.z - M.z*P.y;
        Vb = M.z*P.x - M.x*P.z;
		
		Ha1=Ha*mod1;
		Hb1=Hb*mod2;

		if(false){
		//if (Oc==0 && Hc==0) {// piano orizzontale
			register float Ha2,Hb2;
			--yv;
			if(yv & 1) ++yv;
			while(++yv<=ymax){
				++yv;	// <<<==== da togliere
				for(u=0; u<(pix[yv]-1); ){
				
					p1=px[yv][u];
					p2=px[yv][++u];
					xv=p1.v;
					t=p2.v;
					if (xv<t){
						// uso h come variabili buffer
						h=p2.n-p1.n;  // lunghezza della linea orizontale EFFETTIVA clip escluso
						if (p1.n <p1.v) {	// la texture inizia piu' a sinistra del video 
							a=((float)1/h)*(p1.v-p1.n);
							p1.r=(float)(p2.r-p1.r)*a+p1.r;
							p1.g=(float)(p2.g-p1.g)*a+p1.g;
							p1.b=(float)(p2.b-p1.b)*a+p1.b;
							h=h-(p1.v-p1.n);
						} 
						y=yc_v-yv;
						
						p9=p8+smHWGB5_VOffs*yv+xv;
						--xv;
						x=xv-xc_v;
						
						zpun=(z_buf+smHWGB5_VOffs*yv+xv);

					        
						a =mod1*( Oa + (Ha * x) + (Va * y));
					    b =mod2*( Ob + (Hb * x) + (Vb * y));
						c=Vc*y;
						a/=c;
						b/=c;
						Ha2=Ha1/c;
						Hb2=Hb1/c;
						
						rs=p1.r;ri1=((p2.r-rs)<<col_shc)/h;
						gs=p1.g;gi1=((p2.g-gs)<<col_shc)/h;
						bs=p1.b;bi1=((p2.b-bs)<<col_shc)/h;
						bi2=bs<<col_shc;gi2=gs<<col_shc;ri2=rs<<col_shc;

						dzb_=(float)((norm.x/d)/-cd)*mfl;
						zb_=dzb_*(xv-xc_v)+((float)(( norm.y*(yc_v-yv)/d+norm.z )/-cd) )*mfl;
						
						while(++xv < t) {
			       			zb_+=dzb_;
			         			
							bi2+=bi1;
							gi2+=gi1;						
							ri2+=ri1;

							u1=*zpun;
							v1= u1 & 0x00FFFFFF;	// v1= zbuffer
							u1= u1 & 0xFF000000; 	// u1= frame count
							if (u1 != framet || v1<zb_){
								*zpun=zb_ | framet;
								
				 			        u1 =(long)a &dxt;
							        v1 =(long)b &dyt;

								dat=*(p10+u1+v1); 
								
								bi=(bi2 >> col_sh);
								gi=(gi2 >> col_sh);
								ri=(ri2 >> col_sh);
								
								dat1=dat & mask;
								dat2=*(colb+dat1+bi);
								dat1=(dat>>5) & mask;
								dat2 |= *(colg+dat1+gi);
								dat1=(dat>>10) & mask;
								dat2 |= *(colr+dat1+ri);
								*(p9)=dat2;
							}  ++p9;
							a += Ha2;
						    b += Hb2;
			         		++zpun;

						}	// while (++xv <t)
					}
				}
		
			}
		
		} else {		// Piano qualsiasi
				
			--yv;
			//if(yv & 1) ++yv;
			while(++yv<=ymax){
				//++yv;	// <<<==== da togliere
								
				for(u=0; u<(pix[yv]-1); ){
				
					p1=px[yv][u];
					p2=px[yv][++u];
					xv=p1.v;
					t=p2.v;
					if (xv<t){
						// uso h come variabili buffer
						h=p2.n-p1.n;  // lunghezza della linea orizontale EFFETTIVA clip escluso
						
						if (p1.n <p1.v) {	// la texture inizia piu' a sinistra del video 
							a=((float)1/h)*(p1.v-p1.n);
							p1.r=(float)(p2.r-p1.r)*a+p1.r;
							p1.g=(float)(p2.g-p1.g)*a+p1.g;
							p1.b=(float)(p2.b-p1.b)*a+p1.b;
							h=h-(p1.v-p1.n);
						} 
						y=yc_v-yv;

						p9=p8+smHWGB5_VOffs*yv+xv;
						--xv;
						x=xv-xc_v;

						zpun=(z_buf+smHWGB5_VOffs*yv+xv);

						a =mod1*( Oa + (Ha * x) + (Va * y));
					    b =mod2*( Ob + (Hb * x) + (Vb * y));
					    c =  (Oc + (Hc * x) + (Vc * y));

						rs=p1.r;ri1=( (p2.r-rs)<<col_shc)/h;
						gs=p1.g;gi1=((p2.g-gs)<<col_shc)/h;
						bs=p1.b;bi1=((p2.b-bs)<<col_shc)/h;
						bi2=bs<<col_shc;gi2=gs<<col_shc;ri2=rs<<col_shc;
						dzb_=(float)((norm.x/d)/-cd)*mfl;
						zb_=dzb_*(xv-xc_v)+((float)(( norm.y*(yc_v-yv)/d+norm.z )/-cd) )*mfl;
						while(++xv <= t) {
			         		zb_+=dzb_;
			         			
							bi2+=bi1;
							gi2+=gi1;						
							ri2+=ri1;

							u1=*zpun;
							v1= u1 & 0x00FFFFFF;	// v1= zbuffer
							u1= u1 & 0xFF000000; 	// u1= frame count
							if (u1 != framet || v1<zb_){
								*zpun=zb_ | framet;
								
								
								h=1/c;
							   	u1 =((long)(a *h)) &dxt;
							    v1 =((long)(b *h)) &dyt;
								dat=*(p10+u1+v1); 
								
								bi=(bi2 >> col_sh);
								gi=(gi2 >> col_sh);
								ri=(ri2 >> col_sh);
								
								dat1=dat & mask;
								dat2=*(colb+dat1+bi);
								dat1=(dat>>5) & mask;
								dat2 |= *(colg+dat1+gi);
								dat1=(dat>>10) & mask;
								dat2 |= *(colr+dat1+ri);
								*(p9)=dat2;
							}  ++p9;
							a += Ha1;
					        b += Hb1;
			       			c += Hc;
			       			++zpun;

						}	// while (++xv <t)
					}
				}
		
			}
		}
	}

	
	
	return SM3DErr_NoErr;
}

